from plover_per_application_state.transition.handler.handler import TransitionHandler
from plover_per_application_state.transition.handler.default import DefaultTransitionHandler
from plover_per_application_state.transition.handler.dynamic_title import DynamicTitleTransitionHandler



from plover.engine import StenoEngine
from plover_application_controls.window import WindowTracker

from plover_per_application_state import PerApplicationStateExtension
from plover_per_application_state.transition.handler import DynamicTitleTransitionHandler


def command(engine: StenoEngine, arg: str) -> None:
    subcommand, *args = arg.split(":")
    if not subcommand:
        raise KeyError("No subcommand specified")
    _COMMAND_MAP[subcommand](engine, *args)


def clear(engine: StenoEngine) -> None:
    engine.clear_translator_state()


def clear_all(engine: StenoEngine) -> None:
    PerApplicationStateExtension.get_state_manager().clear()
    engine.clear_translator_state()


def clear_window(engine: StenoEngine) -> None:
    handle_hash = WindowTracker.current_window_details.handle_hash
    PerApplicationStateExtension.get_state_manager().clear_window(handle_hash)
    engine.clear_translator_state()


def prevent_merge(engine: StenoEngine) -> None:
    DynamicTitleTransitionHandler.prevent_merge()


_COMMAND_MAP = {
    "clear": clear,
    "clear_all": clear_all,
    "clear_window": clear_window,
    "prevent_merge": prevent_merge
}



from plover.translation import _State


class WindowStateCollection:
    GLOBAL_LIMIT = (75, 15)
    WINDOW_LIMIT = (25, 5)

    _all_states = {}
    _all_states_count = 0

    _next_timestamp = -1

    def __init__(self, handle_hash: int):
        self._handle_hash = handle_hash
        self._states = {}
        self._state_count = 0

    @staticmethod
    def _timestamp() -> int:
        WindowStateCollection._next_timestamp += 1
        return WindowStateCollection._next_timestamp

    def __contains__(self, title: str) -> bool:
        return title in self._states

    def __getitem__(self, title: str) -> _State:
        try:
            state = self._states[title][0]
            timestamp = WindowStateCollection._timestamp()
            self._states[title] = (state, timestamp)
            global_key = (self._handle_hash, title)
            WindowStateCollection._all_states[global_key] = (self, timestamp)
            return state
        except KeyError:
            return _State()

    def __setitem__(self, title: str, state: _State) -> None:
        timestamp = WindowStateCollection._timestamp()
        new_entry = title not in self._states
        self._states[title] = (state, timestamp)
        global_key = (self._handle_hash, title)
        WindowStateCollection._all_states[global_key] = (self, timestamp)
        if new_entry:
            self._state_count += 1
            WindowStateCollection._all_states_count += 1
        self._check_eviction()
        WindowStateCollection._check_global_eviction()

    def __delitem__(self, title: str) -> None:
        del self._states[title]
        self._state_count -= 1
        global_key = (self._handle_hash, title)
        del WindowStateCollection._all_states[global_key]
        WindowStateCollection._all_states_count -= 1

    def clear(self) -> None:
        for title in self._states:
            global_key = (self._handle_hash, title)
            del WindowStateCollection._all_states[global_key]
        WindowStateCollection._all_states_count -= self._state_count
        self._states = {}
        self._state_count = 0

    def _check_eviction(self) -> None:
        if self._state_count > sum(WindowStateCollection.WINDOW_LIMIT):
            self._evict(WindowStateCollection.WINDOW_LIMIT[1])

    def _evict(self, n: int) -> None:
        candidates = sorted(self._states.items(), key=lambda x: x[1][1])
        for (title, _), _ in zip(candidates, range(n)):
            del self._states[title]
            global_key = (self._handle_hash, title)
            del WindowStateCollection._all_states[global_key]

        self._state_count -= n
        WindowStateCollection._all_states_count -= n

    @staticmethod
    def _check_global_eviction() -> None:
        if WindowStateCollection._all_states_count > sum(WindowStateCollection.GLOBAL_LIMIT):
            WindowStateCollection._global_evict(WindowStateCollection.GLOBAL_LIMIT[1])

    @staticmethod
    def _global_evict(n: int) -> None:
        candidates = sorted(WindowStateCollection._all_states.items(), key=lambda x: x[1][1])
        for ((_, title), (cls, _)), _ in zip(candidates, range(n)):
            del cls[title]



from __future__ import annotations

from plover.engine import StenoEngine

import plover_per_application_state.state.manager as _state_manager
from plover_per_application_state.transition.details import TransitionDetails
from plover_per_application_state.transition.handler.handler import TransitionHandler


class DefaultTransitionHandler(TransitionHandler):
    def handle_transition(self, state_manager: _state_manager.StateManager,
                          engine: StenoEngine,
                          details: TransitionDetails) -> bool:

        stored_state = state_manager.get_state(details.new_details.handle_hash, details.new_details.title)
        state_manager.store_state(details.old_details.handle_hash, details.old_details.title, engine.translator_state)
        engine.translator_state = stored_state
        return True



from __future__ import annotations

from plover.engine import StenoEngine
from plover_application_controls.window import WindowTracker, WindowDetails

from plover_per_application_state.state.manager import StateManager
import plover_per_application_state.transition.handler as handlers
from plover_per_application_state.transition import TransitionDetails


class PerApplicationStateExtension:
    _instance: 'PerApplicationStateExtension' = None

    def __init__(self, engine: StenoEngine) -> None:
        PerApplicationStateExtension._instance = self
        self._on = False
        self._engine = engine

        self._state_manager = StateManager([
            handlers.DynamicTitleTransitionHandler(engine),
            handlers.DefaultTransitionHandler(engine)
        ])
        self._last_details = WindowTracker.BLANK_DETAILS

        WindowTracker.add_callback(True, self.on_window_callback)

    def start(self) -> None:
        self._on = True

    def stop(self) -> None:
        self._on = False

    def on_window_callback(self, _, details: WindowDetails) -> None:
        if not self._on:
            self._last_details = details
            return

        self._state_manager.on_transition(self._engine, TransitionDetails(self._last_details, details))
        self._last_details = details

    @staticmethod
    def get_state_manager() -> StateManager | None:
        if PerApplicationStateExtension._instance is None:
            return None
        return PerApplicationStateExtension._instance._state_manager



from __future__ import annotations

import time

from plover.engine import StenoEngine
from plover.formatting import _Action
from plover_application_controls.window import WindowTracker

import plover_per_application_state.state.manager as _state_manager
from plover_per_application_state.transition.details import TransitionDetails
from plover_per_application_state.transition.handler.handler import TransitionHandler


class DynamicTitleTransitionHandler(TransitionHandler):

    TITLE_CHANGE_TIMEOUT = WindowTracker.CHECK_INTERVAL * 2
    MERGE_PREVENTION_TIMEOUT = TITLE_CHANGE_TIMEOUT

    _last_merge_prevention = 0

    def __init__(self, engine: StenoEngine, priority=100):
        super().__init__(engine, priority)
        self._changes = False
        self._change_timestamp = 0
        engine.hook_connect("translated", self.on_translated)

    def handle_transition(self, state_manager: _state_manager.StateManager,
                          engine: StenoEngine,
                          details: TransitionDetails) -> bool:
        changes = self._changes
        self._changes = False

        if not changes or details.window_changed:
            return False

        current_time = time.time()

        if current_time - DynamicTitleTransitionHandler._last_merge_prevention \
                < DynamicTitleTransitionHandler.MERGE_PREVENTION_TIMEOUT:
            return False

        if current_time - self._change_timestamp \
                > DynamicTitleTransitionHandler.TITLE_CHANGE_TIMEOUT:
            return False

        stored_state = state_manager.get_state(details.new_details.handle_hash, details.new_details.title, False)
        return not stored_state or stored_state == engine.translator_state

    def on_translated(self, old: [_Action], new: [_Action]) -> None:
        for action in new:
            if action.command and action.command.startswith("application:"):
                return

        self._changes = True
        self._change_timestamp = time.time()

    @staticmethod
    def prevent_merge():
        DynamicTitleTransitionHandler._last_merge_prevention = time.time()



from dataclasses import dataclass

from plover_application_controls.window import WindowDetails


@dataclass
class TransitionDetails:
    old_details: WindowDetails
    new_details: WindowDetails

    @property
    def window_changed(self) -> bool:
        return self.old_details.handle_hash != self.new_details.handle_hash



from __future__ import annotations

from plover.engine import StenoEngine
from plover.translation import _State

from plover_per_application_state.state.window_state_collection import WindowStateCollection
import plover_per_application_state.transition.handler.handler as transition_handler
from plover_per_application_state.transition import TransitionDetails


class StateManager:
    WINDOW_LIMIT = (50, 10)

    def __init__(self, handlers: [transition_handler.TransitionHandler]):
        self._window_state_collections = {}
        self._window_state_collection_count = 0
        self._next_timestamp = -1

        self._transition_handlers = handlers
        self._transition_handlers.sort(reverse=True)

    def add_transition_handlers(self, handlers: [transition_handler.TransitionHandler]) -> None:
        self._transition_handlers.extend(handlers)
        self._transition_handlers.sort(reverse=True)

    def remove_transition_handlers(self, handlers: [transition_handler.TransitionHandler]) -> None:
        for handler in handlers:
            self._transition_handlers.remove(handler)

    def on_transition(self, engine: StenoEngine, details: TransitionDetails) -> None:
        for handler in self._transition_handlers:
            if handler.handle_transition(self, engine, details):
                break

    def store_state(self, handle_hash: int, title: str, state: _State) -> None:
        self._get_window_state_collection(handle_hash)[title] = state
        self._check_eviction()

    def get_state(self, handle_hash: int, title: str, default=True) -> _State | None:
        state_collection = self._get_window_state_collection(handle_hash)
        if not default and title not in state_collection:
            return None
        return state_collection[title]

    def _timestamp(self) -> int:
        self._next_timestamp += 1
        return self._next_timestamp

    def _get_window_state_collection(self, handle_hash: int) -> WindowStateCollection:
        timestamp = self._timestamp()
        try:
            state_collection = self._window_state_collections[handle_hash][0]
        except KeyError:
            state_collection = WindowStateCollection(handle_hash)
            self._window_state_collection_count += 1
        self._window_state_collections[handle_hash] = (state_collection, timestamp)
        return state_collection

    def _check_eviction(self) -> None:
        if self._window_state_collection_count > sum(StateManager.WINDOW_LIMIT):
            self._evict(StateManager.WINDOW_LIMIT[1])

    def _evict(self, n: int) -> None:
        candidates = sorted(self._window_state_collections.items(), key=lambda x: x[1][1])
        for (handle_hash, (window_state_collection, _)), _ in zip(candidates, range(n)):
            window_state_collection.clear()
            del self._window_state_collections[handle_hash]
        self._window_state_collection_count -= n

    def clear(self) -> None:
        for window_state_collection, _ in self._window_state_collections.values():
            window_state_collection.clear()
        self._window_state_collections.clear()
        self._window_state_collection_count = 0

    def clear_window(self, handle_hash: int) -> None:
        self._get_window_state_collection(handle_hash).clear()



from __future__ import annotations

import abc
from functools import total_ordering
from typing import final

from plover.engine import StenoEngine

import plover_per_application_state.state.manager as _state_manager
from plover_per_application_state.transition.details import TransitionDetails


@total_ordering
class TransitionHandler(abc.ABC):
    def __init__(self, engine: StenoEngine, priority=0):
        self._engine = engine
        self._priority = priority

    @abc.abstractmethod
    def handle_transition(self, state_manager: _state_manager.StateManager,
                          engine: StenoEngine,
                          details: TransitionDetails) -> bool:
        raise NotImplementedError()

    @final
    def __eq__(self, other: 'TransitionHandler') -> bool:
        return self._priority == other._priority

    @final
    def __gt__(self, other: 'TransitionHandler') -> bool:
        return self._priority > other._priority
